
import re
import hmac
import hashlib
import secrets
import email.utils
from typing import Union, Optional, Tuple, Sequence

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ChallengeResponse)
from ..config import default_config, SASLConfig
from ..creds.client import ClientCredentials
from ..creds.server import ServerCredentials
from ..exception import InvalidResponse, MechanismUnusable, UnexpectedChallenge
from ..identity import Identity

__all__ = ['CramMD5Result', 'CramMD5Mechanism']


class CramMD5Result(ServerCredentials):
    """Because this mechanism uses hash algorithms to compare secrets, the
    :meth:`~CramMD5Mechanism.server_attempt` method returns this sub-class
    which overrides the :meth:`.verify` method.

    """

    __slots__: Sequence[str] = ['_username', '_challenge', '_digest',
                                '_config']

    def __init__(self, username: str, challenge: bytes,
                 digest: bytes, *, config: SASLConfig) -> None:
        super().__init__()
        self._username = username
        self._challenge = challenge
        self._digest = digest
        self._config = config

    @property
    def authcid(self) -> str:
        return self._username

    @property
    def authzid(self) -> str:
        return self._username

    def verify(self, identity: Optional[Identity]) -> bool:
        if identity is None:
            return False
        clear_secret = identity.get_clear_secret()
        if clear_secret is None:
            raise MechanismUnusable('CRAM-MD5')
        secret_b = clear_secret.encode('utf-8')
        expected_hmac = hmac.new(secret_b, self._challenge, hashlib.md5)
        expected_digest = expected_hmac.hexdigest().encode('ascii')
        prepare = self._config.prepare
        self_authcid = prepare(self.authcid)
        other_authcid = prepare(identity.authcid)
        return secrets.compare_digest(self_authcid, other_authcid) \
            and hmac.compare_digest(expected_digest, self._digest)


class CramMD5Mechanism(ServerMechanism, ClientMechanism):
    """Implements the CRAM-MD5 authentication mechanism.

    Warning:
        Although secure during transport, offering this mechanism can be
        dangerous, as it can have implications about how the credentials are
        stored server-side.

    """

    _pattern = re.compile(br'^(.*) ([^ ]+)$')

    def __init__(self, name: Union[str, bytes] = b'CRAM-MD5',
                 config: SASLConfig = default_config) -> None:
        super().__init__(name, config)

    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[CramMD5Result, None]:
        try:
            first = responses[0]
        except IndexError as exc:
            challenge = email.utils.make_msgid().encode('utf-8')
            raise ServerChallenge(challenge) from exc

        match = re.match(self._pattern, first.response)
        if not match:
            raise InvalidResponse()
        username, digest = match.groups()

        username_str = username.decode('utf-8')
        result = CramMD5Result(username_str, first.challenge, digest,
                               config=self.config)
        return result, None

    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        if len(challenges) < 1:
            return ChallengeResponse(b'', b'')
        elif len(challenges) > 1:
            raise UnexpectedChallenge()
        challenge = challenges[0].data
        prepare = self.config.prepare
        authcid = prepare(creds.authcid).encode('utf-8')
        secret = prepare(creds.secret).encode('utf-8')
        digest = hmac.new(secret, challenge, hashlib.md5).hexdigest()
        response = b' '.join((authcid, digest.encode('ascii')))
        return ChallengeResponse(challenge, response)




import re
from typing import Union, Tuple, Sequence

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ChallengeResponse)
from ..config import default_config, SASLConfig
from ..creds.client import ClientCredentials
from ..creds.plain import PlainCredentials
from ..exception import InvalidResponse, UnexpectedChallenge

__all__ = ['PlainMechanism']


class PlainMechanism(ServerMechanism, ClientMechanism):
    """Implements the PLAIN authentication mechanism."""

    _pattern = re.compile(br'^([^\x00]*)\x00([^\x00]+)\x00([^\x00]*)$')

    __slots__: Sequence[str] = []

    def __init__(self, name: Union[str, bytes] = b'PLAIN',
                 config: SASLConfig = default_config) -> None:
        super().__init__(name, config)

    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[PlainCredentials, None]:
        try:
            first = responses[0]
        except IndexError as exc:
            raise ServerChallenge(b'') from exc

        match = re.match(self._pattern, first.response)
        if not match:
            raise InvalidResponse()
        zid, cid, secret = match.groups()

        cid_str = cid.decode('utf-8')
        secret_str = secret.decode('utf-8')
        zid_str = zid.decode('utf-8') or cid_str
        return PlainCredentials(cid_str, secret_str, zid_str), None

    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        if len(challenges) == 0:
            challenge = b''
        elif len(challenges) == 1:
            challenge = challenges[0].data
        else:
            raise UnexpectedChallenge()
        authzid = creds.authzid.encode('utf-8')
        authcid = creds.authcid.encode('utf-8')
        secret = creds.secret.encode('utf-8')
        response = b'\0'.join((authzid, authcid, secret))
        return ChallengeResponse(challenge, response)




from __future__ import absolute_import

import unittest

from pysasl import SASLAuth
from pysasl.creds.client import ClientCredentials
from pysasl.creds.external import (ExternalVerificationRequired,
                                   ExternalCredentials)
from pysasl.exception import InvalidResponse, UnexpectedChallenge
from pysasl.identity import ClearIdentity
from pysasl.mechanism import ServerChallenge, ChallengeResponse
from pysasl.mechanism.oauth import OAuth2Mechanism


class TestOAuth2Mechanism(unittest.TestCase):

    def setUp(self) -> None:
        self.mech = OAuth2Mechanism()

    def test_availability(self) -> None:
        sasl = SASLAuth.defaults()
        self.assertIsNone(sasl.get_server(b'XOAUTH2'))
        self.assertIsNone(sasl.get_client(b'XOAUTH2'))
        sasl = SASLAuth.named([b'XOAUTH2'])
        self.assertEqual(self.mech, sasl.get_server(b'XOAUTH2'))
        self.assertEqual(self.mech, sasl.get_client(b'XOAUTH2'))
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)

    def test_server_attempt_issues_challenge(self) -> None:
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'', exc.data)
        else:
            self.fail('ServerChallenge not raised')

    def test_server_attempt_bad_response(self) -> None:
        self.assertRaises(InvalidResponse,
                          self.mech.server_attempt,
                          [ChallengeResponse(b'', b'abcdefghi')])

    def test_server_attempt_successful(self) -> None:
        result, final = self.mech.server_attempt([ChallengeResponse(
            b'', b'user=testuser\x01auth=Bearer testtoken\x01\x01')])
        self.assertIsNone(final)
        self.assertIsInstance(result, ExternalCredentials)
        self.assertEqual('', result.authcid)
        self.assertEqual('testuser', result.authzid)
        with self.assertRaises(ExternalVerificationRequired) as exc:
            result.verify(ClearIdentity('testuser', 'secret'))
        self.assertEqual('testtoken', exc.exception.token)
        with self.assertRaises(ExternalVerificationRequired) as exc:
            result.verify(None)
        self.assertEqual('testtoken', exc.exception.token)

    def test_client_attempt(self) -> None:
        creds = ClientCredentials('testuser', 'testtoken')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'user=testuser\x01auth=Bearer testtoken\x01\x01',
                         resp1.response)
        resp2 = self.mech.client_attempt(creds, [ServerChallenge(b'')])
        self.assertEqual(b'user=testuser\x01auth=Bearer testtoken\x01\x01',
                         resp2.response)
        self.assertRaises(UnexpectedChallenge,
                          self.mech.client_attempt,
                          creds, [ServerChallenge(b'')]*2)

    def test_client_attempt_error(self) -> None:
        creds = ClientCredentials('testuser', 'testtoken')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'user=testuser\x01auth=Bearer testtoken\x01\x01',
                         resp1.response)
        resp2 = self.mech.client_attempt(creds, [
            ServerChallenge(b'{"status":"401","schemes":"bearer mac",'
                            b'"scope":"https://mail.google.com/"}\n')])
        self.assertEqual(b'', resp2.response)




from typing import Optional

from . import Credentials

__all__ = ['ClientCredentials']


class ClientCredentials(Credentials):
    """Credentials that are provided by the user and transmitted to the server
    for authentication..

    Args:
        authcid: The authentication identity, e.g. a login username.
        secret: The secret string, e.g. password.
        authzid: The authorization identity, or an empty string.

    """

    __slots__ = ['_authcid', '_secret', '_authzid']

    def __init__(self, authcid: str, secret: str,
                 authzid: Optional[str] = None) -> None:
        super().__init__()
        self._authcid = authcid
        self._secret = secret
        self._authzid = authzid or ''

    @property
    def authcid(self) -> str:
        return self._authcid

    @property
    def secret(self) -> str:
        """The secret string, e.g. password."""
        return self._secret

    @property
    def authzid(self) -> str:
        return self._authzid

    def __repr__(self) -> str:
        return f'ClientCredentials({self.authcid!r}, ..., {self.authzid!r})'



# type: ignore

import os
import os.path
from invoke import task, Collection

from . import check, doc, lint, test, types

try:
    from shlex import join
except ImportError:  # Python < 3.8
    def join(args):
        return ' '.join(args)


@task
def clean(ctx, full=False):
    """Delete all the standard build and validate artifacts."""
    if full:
        ctx.run('git clean -dfX')
    else:
        anywhere = ['__pycache__']
        top_level = [
            '.coverage',
            '.mypy_cache',
            '.pytest_cache',
            'dist',
            'doc/build/']
        for name in anywhere:
            for path in [ctx.package, 'test']:
                subpaths = [os.path.join(subpath, name)
                            for subpath, dirs, names in os.walk(path)
                            if name in dirs or name in names]
                for subpath in subpaths:
                    ctx.run(join(['rm', '-rf', subpath]))
        for name in top_level:
            ctx.run(join(['rm', '-rf', name]))


@task
def install(ctx, dev=True, update=False):
    """Install the library and all development tools."""
    choice = 'dev' if dev else 'all'
    if update:
        ctx.run('pip install -U -r requirements-{}.txt'.format(choice))
    else:
        ctx.run('pip install -r requirements-{}.txt'.format(choice))


@task(test.all, types.all, lint.all)
def validate(ctx):
    """Run all tests, type checks, and linters."""
    pass


ns = Collection(clean, install)
ns.add_task(validate, default=True)
ns.add_collection(check)
ns.add_collection(test)
ns.add_collection(types)
ns.add_collection(lint)
ns.add_collection(doc)

ns.configure({
    'package': 'pysasl',
    'run': {
        'echo': True,
        'pty': True,
    }
})




import secrets
from typing import Optional

from .server import ServerCredentials
from ..identity import Identity
from ..prep import default_prep, Preparation

__all__ = ['PlainCredentials']


class PlainCredentials(ServerCredentials):
    """Implementation of :class:`~pysasl.creds.server.ServerCredentials` for
    typical SASL mechanisms like
    :class:`~pysasl.mechanism.plain.PlainMechanism` where the mechanism
    operates on the *secret* string in cleartext.

    Args:
        authcid: Authentication ID string (the username).
        secret: Secret string (the password).
        authzid: Authorization ID string, if provided.
        prepare: The preparation algorithm function.

    """

    __slots__ = ['_authcid', '_secret', '_authzid', '_prepare']

    def __init__(self, authcid: str, secret: str, authzid: str = '', *,
                 prepare: Preparation = default_prep) -> None:
        super().__init__()
        self._authcid = authcid
        self._secret = secret
        self._authzid = authzid or prepare(authcid)
        self._prepare = prepare

    @property
    def authcid(self) -> str:
        return self._authcid

    @property
    def authzid(self) -> str:
        return self._authzid

    def verify(self, identity: Optional[Identity]) -> bool:
        if identity is not None:
            prepare = self._prepare
            self_authcid = prepare(self.authcid)
            other_authcid = prepare(identity.authcid)
            return secrets.compare_digest(self_authcid, other_authcid)  \
                and identity.compare_secret(self._secret)
        return False

    def __repr__(self) -> str:
        return f'PlainCredentials({self.authcid!r}, ..., {self.authzid!r})'



# type: ignore

from invoke import task, Collection

from .check import check_import


@task(check_import)
def pytest(ctx):
    """Run the unit tests with py.test."""
    ctx.run('py.test --cov={} --cov-report=term-missing'.format(ctx.package))


@task(pytest)
def all(ctx):
    """Run all test utilities."""
    pass


ns = Collection(pytest)
ns.add_task(all, default=True)




import warnings
from abc import abstractmethod
from typing_extensions import Final, Protocol

try:
    from passlib.utils import saslprep as _saslprep
except ImportError as exc:  # pragma: no cover
    _saslprep = None
    _saslprep_exc = exc
    warnings.warn('passlib.utils.saslprep is not available', ImportWarning)

__all__ = ['Preparation', 'default_prep', 'noprep', 'saslprep']


class Preparation(Protocol):
    """Any callable that prepares a string value to improve the likelihood that
    comparisons behave in an expected manner.

    See Also:
        `RFC 4422 5.
        <https://datatracker.ietf.org/doc/html/rfc4422#section-5>`_

    """

    @abstractmethod
    def __call__(self, source: str) -> str:
        ...


def noprep(source: str) -> str:  # pragma: no cover
    """A :class:`Preparation` implementation that returns the *source* value
    unchanged.

    Args:
        source: The string to prepare.

    """
    return source


def saslprep(source: str) -> str:
    """The SASLprep algorithm defined by `RFC 4013
    <https://datatracker.ietf.org/doc/html/rfc4013>`_, implemented by
    :func:`passlib.utils.saslprep`.

    Args:
        source: The string to prepare.

    Raises:
        ImportError: The implementation is not available.

    """
    if _saslprep is not None:
        ret: str = _saslprep(source)
        return ret
    else:  # pragma: no cover
        raise _saslprep_exc


if _saslprep is not None:
    _default_prep = saslprep
else:  # pragma: no cover
    _default_prep = noprep

#: Prepares the *source* string using the default preparation algorithm. This
#: default is :func:`saslprep` if available otherwise :func:`noprep`.
default_prep: Final = _default_prep




from dataclasses import dataclass
from typing_extensions import Final

from .prep import default_prep, Preparation

__all__ = ['SASLConfig', 'default_config']


@dataclass(frozen=True)
class SASLConfig:
    """Provides any configuration necessary for
    :class:`~pysasl.mechanism.ServerMechanism` or
    :class:`~pysasl.mechanism.ClientMechanism` instances.

    """

    prepare: Preparation = default_prep
    """The preparation algorithm function."""


@dataclass(frozen=True, init=False, repr=False)
class _DefaultConfig(SASLConfig):

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return 'SASLConfig()'


#: A configuration instance with all defaults.
default_config: Final = _DefaultConfig()





from abc import abstractmethod, ABCMeta
from typing import Union, Optional, Tuple, Sequence
from typing_extensions import TypeAlias

from ..config import SASLConfig
from ..creds.client import ClientCredentials
from ..creds.server import ServerCredentials

__all__ = ['Mechanism', 'ServerChallenge', 'ChallengeResponse',
           'ServerMechanism', 'ClientMechanism']

#: A type alias for either server or client mechanisms.
Mechanism: TypeAlias = Union['ServerMechanism', 'ClientMechanism']


class ServerChallenge(Exception):
    """Raised by :meth:`~ServerMechanism.server_attempt` to provide server
    challenges.

    Args:
        data: The challenge string that should be sent to the client.

    """

    __slots__ = ['_data']

    def __init__(self, data: bytes) -> None:
        super().__init__(data)
        self._data = data

    @property
    def data(self) -> bytes:
        """The server challenge that should be sent to the client."""
        return self._data

    def __repr__(self) -> str:
        return f'ServerChallenge({self.data!r})'


class ChallengeResponse:
    """A challenge-response exchange between server and client.

    Args:
        challenge: The server challenge string.
        response: The client response string.

    """

    __slots__ = ['_challenge', '_response']

    def __init__(self, challenge: bytes, response: bytes) -> None:
        super().__init__()
        self._challenge = challenge
        self._response = response

    @property
    def challenge(self) -> bytes:
        """The server challenge string."""
        return self._challenge

    @property
    def response(self) -> bytes:
        """The client response string."""
        return self._response

    def __repr__(self) -> str:
        return f'ChallengeResponse({self.challenge!r}, {self.response!r})'


class _BaseMechanism:

    __slots__: Sequence[str] = ['_name', '_config']

    def __init__(self, name: Union[str, bytes], config: SASLConfig) -> None:
        super().__init__()
        if isinstance(name, str):
            name = name.encode('ascii')
        self._name = name
        self._config = config

    @property
    def name(self) -> bytes:
        """The SASL name for this mechanism."""
        return self._name

    @property
    def config(self) -> SASLConfig:
        """The configuration object."""
        return self._config

    def __eq__(self, other: object) -> bool:
        if isinstance(other, _BaseMechanism):
            return self.name == other.name and self.config == other.config
        return NotImplemented


class ServerMechanism(_BaseMechanism, metaclass=ABCMeta):
    """Base class for implementing SASL mechanisms that support server-side
    credential verification.

    """

    __slots__: Sequence[str] = []

    @abstractmethod
    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[ServerCredentials, Optional[bytes]]:
        """For SASL server-side credential verification, receives responses
        from the client and issues challenges until it has everything needed to
        verify the credentials.

        If a challenge is necessary, a :class:`ServerChallenge` exception will
        be raised. The response to this challenge must then be added to
        *responses* in the next call to :meth:`.server_attempt`.

        Args:
            responses: The challenge-response exchanges thus far.

        Returns:
            A tuple of the authentication credentials received from the client
            once no more challenges are necessary, and an optional final
            response string from the server used by some mechanisms.

        Raises:
            ServerChallenge: The server challenge needing a client response.
            InvalidResponse: The server received an invalid client response.

        """
        ...


class ClientMechanism(_BaseMechanism, metaclass=ABCMeta):
    """Base class for implementing SASL mechanisms that support client-side
    credential verification.

    """

    __slots__: Sequence[str] = []

    @abstractmethod
    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        """For SASL client-side credential verification, produce responses to
        send to the server and react to its challenges until the server returns
        a final success or failure.

        Args:
            creds: The credentials to attempt authentication with.
            challenges: The server challenges received.

        Returns:
            The response to the most recent server challenge.

        Raises:
            UnexpectedChallenge: The server has issued a challenge the client
                mechanism does not recognize.

        """
        ...



# type: ignore

from invoke import task, Collection

from .check import check_import


@task(check_import)
def flake8(ctx):
    """Run the flake8 linter."""
    ctx.run('flake8 {} test {} *.py'.format(ctx.package, __package__))


@task(check_import)
def bandit(ctx):
    """Run the bandit linter."""
    ctx.run('bandit -c pyproject.toml -qr {}'.format(ctx.package))


@task(flake8, bandit)
def all(ctx):
    """Run all linters."""
    pass


ns = Collection(flake8, bandit)
ns.add_task(all, default=True)



# type: ignore

import warnings

from invoke import task, Collection


@task
def check_import(ctx):
    """Check that the library can be imported."""
    try:
        __import__(ctx.package)
    except ImportError:
        warnings.warn('Could not import {!r}, '
                      'task may fail'.format(ctx.package))


ns = Collection()
ns.add_task(check_import, default=True)




import secrets
from abc import abstractmethod
from typing import Optional, Sequence
from typing_extensions import Protocol

from .hashing import HashInterface, Cleartext
from .prep import default_prep, Preparation

__all__ = ['Identity', 'ClearIdentity', 'HashedIdentity']


class Identity(Protocol):
    """Represents an server-side identity that credentials will be
    authenticated against.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def authcid(self) -> str:
        """The authentication identity, e.g. a login username."""
        ...

    @abstractmethod
    def compare_secret(self, secret: str) -> bool:
        """Compare the identity's secret with the given *secret*. The
        comparison must account for things like hashing or token algorithms.

        Args:
            secret: The authentication secret string value.

        """
        ...

    @abstractmethod
    def get_clear_secret(self) -> Optional[str]:
        """Return the cleartext secret string if it is available. This value
        has already been prepared with a :class:`~pysasl.prep.Preparation`
        function.

        """
        ...


class ClearIdentity(Identity):
    """An :class:`Identity` that stores the secret string in cleartext.

    Args:
        authcid: The authentication identity, e.g. a login username.
        secret: The cleartext secret string.
        prepare: The preparation algorithm function.

    """

    __slots__: Sequence[str] = ['_authcid', '_secret', '_prepare']

    def __init__(self, authcid: str, secret: str, *,
                 prepare: Preparation = default_prep) -> None:
        super().__init__()
        self._authcid = authcid
        self._secret = prepare(secret)
        self._prepare = prepare

    @property
    def authcid(self) -> str:
        return self._authcid

    def compare_secret(self, secret: str) -> bool:
        return secrets.compare_digest(self._secret, self._prepare(secret))

    def get_clear_secret(self) -> str:
        """Return the cleartext secret string."""
        return self._secret

    def __repr__(self) -> str:
        return f'ClearIdentity({self.authcid!r}, ...)'


class HashedIdentity(Identity):
    """An :class:`Identity` where the secret has been hashed for storage.

    Args:
        authcid: The authentication identity, e.g. a login username.
        digest: The hashed secret string, using :attr:`.hash`.
        hash: The hash algorithm to use to verify the secret.
        prepare: The preparation algorithm function.

    """

    __slots__: Sequence[str] = ['_authcid', '_digest', '_hash', '_prepare']

    def __init__(self, authcid: str, digest: str, *,
                 hash: HashInterface,
                 prepare: Preparation = default_prep) -> None:
        super().__init__()
        self._authcid = authcid
        self._digest = digest
        self._hash = hash
        self._prepare = prepare

    @classmethod
    def create(cls, authcid: str, secret: str, *,
               hash: HashInterface,
               prepare: Preparation = default_prep) -> 'HashedIdentity':
        """Prepare and hash the given *secret*, returning a
        :class:`HashedIdentity`.

        Args:
            authcid: The authentication identity, e.g. a login username.
            secret: The cleartext secret string.
            hash: The hash algorithm to use to verify the secret.
            prepare: The preparation algorithm function.

        """
        digest = hash.hash(prepare(secret))
        return cls(authcid, digest, hash=hash, prepare=prepare)

    @property
    def authcid(self) -> str:
        return self._authcid

    @property
    def digest(self) -> str:
        """The hashed secret string, using :attr:`.hash`."""
        return self._digest

    @property
    def hash(self) -> HashInterface:
        """The hash implementation to use to verify the secret."""
        return self._hash

    def compare_secret(self, secret: str) -> bool:
        return self._hash.verify(self._prepare(secret), self._digest)

    def get_clear_secret(self) -> Optional[str]:
        """Return the cleartext secret string, only if :attr:`.hash` is
        :class:`~pysasl.hashing.Cleartext`.

        """
        if isinstance(self.hash, Cleartext):
            return self.digest
        else:
            return None

    def __repr__(self) -> str:
        return f'HashedIdentity({self.authcid}, ..., hash={self._hash!r})'




from typing import Optional, Sequence, NoReturn
from typing_extensions import Final

from .server import ServerCredentials
from ..exception import AuthenticationError
from ..identity import Identity

__all__ = ['ExternalVerificationRequired', 'ExternalCredentials']


class ExternalVerificationRequired(AuthenticationError):
    """The credentials are structurally valid but require external
    verification.

    If *token* is ``None``, the credentials provided no additional information
    for verification. Otherwise, *token* should be verified and authorized for
    *identity*.

    Args:
        identity: The identity resolved from the credentials.
        token: A bearer token, if required for verification.

    """

    __slots__: Sequence[str] = ['identity', 'token']

    def __init__(self, identity: Optional[Identity],
                 token: Optional[str] = None) -> None:
        super().__init__()
        self.identity: Final = identity
        self.token: Final = token


class ExternalCredentials(ServerCredentials):
    """Credentials that require external verification, rather than by a
    traditional hashing algorithm.

    Args:
        authzid: Authorization ID string.
        token: A bearer token, if required for verification.

    """

    __slots__: Sequence[str] = ['_authzid', '_token']

    def __init__(self, authzid: str, token: Optional[str] = None) -> None:
        super().__init__()
        self._authzid = authzid
        self._token = token

    @property
    def authcid(self) -> str:
        return ''

    @property
    def authzid(self) -> str:
        return self._authzid

    def verify(self, identity: Optional[Identity]) -> NoReturn:
        """This method always throws :exc:`ExternalVerificationRequired`. For
        applications to support these types of credentials, they must catch
        this exception and use it to authenticate and authorize the request.

        Args:
            identity: The identity being authenticated.

        Raises:
            ExternalVerificationRequired: Always thrown.

        """
        raise ExternalVerificationRequired(identity, self._token)

    def __repr__(self) -> str:
        return f'ExternalCredentials({self.authzid}, ...)'




from __future__ import absolute_import

import unittest

from pysasl import SASLAuth
from pysasl.creds.client import ClientCredentials
from pysasl.creds.plain import PlainCredentials
from pysasl.exception import UnexpectedChallenge
from pysasl.identity import ClearIdentity
from pysasl.mechanism import ServerChallenge, ChallengeResponse
from pysasl.mechanism.login import LoginMechanism


class TestLoginMechanism(unittest.TestCase):

    def setUp(self) -> None:
        self.mech = LoginMechanism()

    def test_availability(self) -> None:
        sasl = SASLAuth.defaults()
        self.assertEqual(self.mech, sasl.get_server(b'LOGIN'))
        self.assertEqual(self.mech, sasl.get_client(b'LOGIN'))
        sasl = SASLAuth.named([b'LOGIN'])
        self.assertEqual(self.mech, sasl.get_server(b'LOGIN'))
        self.assertEqual(self.mech, sasl.get_client(b'LOGIN'))
        sasl = SASLAuth([self.mech])
        self.assertEqual([self.mech], sasl.client_mechanisms)
        self.assertEqual([self.mech], sasl.server_mechanisms)

    def test_server_attempt_issues_challenges(self) -> None:
        try:
            self.mech.server_attempt([])
        except ServerChallenge as exc:
            self.assertEqual(b'Username:', exc.data)
        else:
            self.fail('first ServerChallenge not raised')
        try:
            self.mech.server_attempt([
                ChallengeResponse(b'', b'test')])
        except ServerChallenge as exc:
            self.assertEqual(b'Password:', exc.data)
        else:
            self.fail('second ServerChallenge not raised')

    def test_server_attempt_successful(self) -> None:
        result, final = self.mech.server_attempt([
            ChallengeResponse(b'Username:', b'testuser'),
            ChallengeResponse(b'Password:', b'testpass')])
        self.assertIsNone(final)
        self.assertIsInstance(result, PlainCredentials)
        self.assertEqual('testuser', result.authcid)
        self.assertEqual('testuser', result.authzid)
        self.assertTrue(result.verify(ClearIdentity('testuser', 'testpass')))
        self.assertFalse(result.verify(ClearIdentity('testuser', 'badpass')))
        self.assertFalse(result.verify(ClearIdentity('baduser', 'testpass')))

    def test_client_attempt(self) -> None:
        creds = ClientCredentials('testuser', 'testpass')
        resp1 = self.mech.client_attempt(creds, [])
        self.assertEqual(b'', resp1.response)
        resp2 = self.mech.client_attempt(creds, [
            ServerChallenge(b'Username:')])
        self.assertEqual(b'testuser', resp2.response)
        resp3 = self.mech.client_attempt(creds, [
            ServerChallenge(b'Username:'), ServerChallenge(b'Password:')])
        self.assertEqual(b'testpass', resp3.response)
        self.assertRaises(UnexpectedChallenge,
                          self.mech.client_attempt,
                          creds, [
                              ServerChallenge(b'Username:'),
                              ServerChallenge(b'Password:'),
                              ServerChallenge(b'')])




from typing import Union, Tuple, Sequence

from . import (ServerMechanism, ClientMechanism, ServerChallenge,
               ChallengeResponse)
from ..config import default_config, SASLConfig
from ..creds.client import ClientCredentials
from ..creds.external import ExternalCredentials
from ..exception import UnexpectedChallenge

__all__ = ['ExternalMechanism']


class ExternalMechanism(ServerMechanism, ClientMechanism):
    """Implements the EXTERNAL authentication mechanism.

    See Also:
        `RFC 4422 Appendix A <https://tools.ietf.org/html/rfc4422#appendix-A>`_

    """

    def __init__(self, name: Union[str, bytes] = b'EXTERNAL',
                 config: SASLConfig = default_config) -> None:
        super().__init__(name, config)

    def server_attempt(self, responses: Sequence[ChallengeResponse]) \
            -> Tuple[ExternalCredentials, None]:
        try:
            first = responses[0]
        except IndexError as exc:
            raise ServerChallenge(b'') from exc
        authzid_str = first.response.decode('utf-8')
        return ExternalCredentials(authzid_str), None

    def client_attempt(self, creds: ClientCredentials,
                       challenges: Sequence[ServerChallenge]) \
            -> ChallengeResponse:
        if len(challenges) == 0:
            challenge = b''
        elif len(challenges) == 1:
            challenge = challenges[0].data
        else:
            raise UnexpectedChallenge()
        authzid = creds.authzid.encode('utf-8')
        return ChallengeResponse(challenge, authzid)




from collections import OrderedDict
from typing import Optional, Iterable, Sequence
from typing_extensions import Final

import pkg_resources

from . import mechanism
from .config import default_config, SASLConfig
from .mechanism import Mechanism, ServerMechanism, ClientMechanism

__all__ = ['__version__', 'SASLConfig', 'SASLAuth']

#: The pysasl package version.
__version__: str = pkg_resources.require(__package__)[0].version


class SASLAuth(SASLConfig):
    """Manages the mechanisms available for authentication attempts.

    Args:
        mechanisms: List of available SASL mechanism objects.
        config: The configuration object.

    """

    __slots__ = ['config', '_server_mechanisms', '_client_mechanisms']

    def __init__(self, mechanisms: Sequence[Mechanism], *,
                 config: SASLConfig = default_config) -> None:
        super().__init__()
        self.config: Final = config
        self._server_mechanisms = OrderedDict(
            (mech.name, mech)
            for mech in mechanisms if isinstance(mech, ServerMechanism))
        self._client_mechanisms = OrderedDict(
            (mech.name, mech)
            for mech in mechanisms if isinstance(mech, ClientMechanism))

    @classmethod
    def defaults(cls, *, config: SASLConfig = default_config) -> 'SASLAuth':
        """Uses the default built-in authentication mechanisms, ``PLAIN`` and
        ``LOGIN``.

        Args:
            config: The configuration object.

        Returns:
            A new :class:`SASLAuth` object.

        """
        return cls.named([b'PLAIN', b'LOGIN'])

    @classmethod
    def named(cls, names: Iterable[bytes], *,
              config: SASLConfig = default_config) -> 'SASLAuth':
        """Uses the built-in authentication mechanisms that match a provided
        name.

        Args:
            names: The authentication mechanism names.
            config: The configuration object.

        Returns:
            A new :class:`SASLAuth` object.

        Raises:
            KeyError: A mechanism name was not recognized.

        """
        builtin = {m.name: m for m in cls._get_builtin_mechanisms(config)}
        return SASLAuth([builtin[name] for name in names])

    @classmethod
    def _get_builtin_mechanisms(cls, config: SASLConfig) \
            -> Iterable[Mechanism]:
        group = mechanism.__package__
        for entry_point in pkg_resources.iter_entry_points(group):
            mech_cls = entry_point.load()
            yield mech_cls(entry_point.name, config)

    @property
    def server_mechanisms(self) -> Sequence[ServerMechanism]:
        """List of available :class:`ServerMechanism` objects."""
        return list(self._server_mechanisms.values())

    @property
    def client_mechanisms(self) -> Sequence[ClientMechanism]:
        """List of available :class:`ClientMechanism` objects."""
        return list(self._client_mechanisms.values())

    def get_server(self, name: bytes) -> Optional[ServerMechanism]:
        """Like :meth:`.get`, but only mechanisms inheriting
        :class:`ServerMechanism` will be returned.

        Args:
            name: The SASL mechanism name.

        Returns:
            The mechanism object or ``None``

        """
        return self._server_mechanisms.get(name.upper())

    def get_client(self, name: bytes) -> Optional[ClientMechanism]:
        """Like :meth:`.get`, but only mechanisms inheriting
        :class:`ClientMechanism` will be returned.

        Args:
            name: The SASL mechanism name.

        Returns:
            The mechanism object or ``None``

        """
        return self._client_mechanisms.get(name.upper())




from abc import abstractmethod
from typing import Sequence
from typing_extensions import Protocol

__all__ = ['Credentials']


class Credentials(Protocol):
    """SASL authentication credentials consist of an authentication identity
    and an authorization identity, the identity to be assumed.

    Consider a UNIX system where ``root`` is the superuser and only it may
    assume the identity of other users. With an authentication identity of
    ``root`` and an authorization identity of ``terry``, the authorization
    would succeed because the authentication identity has sufficient
    privileges to assume the authorization identity. If the authentication
    identity were ``greg``, authorization would fail because ``greg`` does not
    have superuser privileges to assume the identity of ``terry``.

    See:
        :class:`~pysasl.creds.identity.Identity`,
        `RFC 4422 2. <https://www.rfc-editor.org/rfc/rfc4422#section-2>`_

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def authcid(self) -> str:
        """The authentication identity, e.g. a login username."""
        ...

    @property
    @abstractmethod
    def authzid(self) -> str:
        """The authorization identity. The :attr:`.authcid` identity must have
        sufficient privileges to assume this identity for the authentication
        attempt to succeed.

        """
        ...
